#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 10:49:45 2019

Tests for visualization.py

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import pytest

# custom modules
import fiducia.visualization as vis